/*
 * Copyright (c) 2019 Hanspeter Portner (dev@open-music-kontrollers.ch)
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the Artistic License 2.0 as published by
 * The Perl Foundation.
 *
 * This source is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * Artistic License 2.0 for more details.
 *
 * You should have received a copy of the Artistic License 2.0
 * along the source as a COPYING file. If not, obtain it from
 * http://www.perlfoundation.org/artistic_license_2_0.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inttypes.h>
#include <dlfcn.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <regex.h>
#include <setjmp.h>

#include <jit_lv2.h>
#include <props.h>
#include <varchunk.h>

#include <jit_api.h>

#include "mum.h"

#define MAX_CHANNEL 2

typedef union _hash_t hash_t;
typedef struct _voice_t voice_t;
typedef struct _dsp_t dsp_t;
typedef struct _job_t job_t;
typedef struct _plughandle_t plughandle_t;

typedef enum _voice_state_t {
	VOICE_STATE_INACTIVE     = 0,
	VOICE_STATE_ACTIVE       = (1 << 0),
	VOICE_STATE_SUSTAIN      = (1 << 1),
	VOICE_STATE_DEACTIVATING = (1 << 2)
} voice_state_t;

union _hash_t {
	struct {
		uint8_t chn;
		uint8_t key;
	};
	uint16_t id;
};

struct _voice_t {
	float gate;
	float gain;
	float freq;
	float pressure;
	float timbre;
	voice_state_t state;
	int32_t remaining;
	hash_t hash;
};

struct _dsp_t {
	plughandle_t *handle;
	char template_src [32];
	char template_so [32];
	void *lib;
	jit_desc_t const *desc;
	varchunk_t *to_worker;
	varchunk_t *to_dsp;
	void *data;
	voice_t voices [JIT_MAX_VOICES];
};

typedef enum _job_type_t {
	JOB_TYPE_DSP_NEW,
	JOB_TYPE_DSP_SET,
	JOB_TYPE_DSP_FREE,
	JOB_TYPE_ERROR_CLEAR,
	JOB_TYPE_ERROR_APPEND,
	JOB_TYPE_ERROR_FREE,
	JOB_TYPE_WORK
} job_type_t;

struct _job_t {
	job_type_t type;
	union {
		dsp_t *dsp;
		char *error;
	};
};

struct _plughandle_t {
	LV2_URID_Map *map;
	LV2_Worker_Schedule *sched;
	LV2_Atom_Forge forge;
	LV2_Atom_Forge_Ref ref;

	LV2_Log_Log *log;
	LV2_Log_Logger logger;

	LV2_URID midi_MidiEvent;

	plugstate_t state;
	plugstate_t stash;

	const LV2_Atom_Sequence *control;
	LV2_Atom_Sequence *notify;
	const float *audio_in [MAX_CHANNEL];
	float *audio_out [MAX_CHANNEL];
	unsigned nchannel;

	PROPS_T(props, MAX_NPROPS);

	varchunk_t *to_loader;

	uint32_t xfade_max;
	uint32_t xfade_cur;

	uint32_t srate;
	char bundle_path [PATH_MAX];

	LV2_URID jit_error;
	bool dirty;

	bool play;
	dsp_t *dsp [2];

	uint16_t rpn_lsb [0x10];
	uint16_t rpn_msb [0x10];
	uint16_t data_lsb [0x10];
	uint16_t gain [0x10];
	uint16_t pressure [0x10];
	uint16_t timbre [0x10];
	float bend [0x10];
	float range [0x10];
	bool sustain [0x10];
};

static int
_vprintf(plughandle_t *handle, LV2_URID type, const char *fmt, va_list ap) 
{
	if(handle->log)
	{
		return lv2_log_vprintf(&handle->logger, type, fmt, ap);
	}

	return 1;
}

static int
_printf(plughandle_t *handle, LV2_URID type, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);

	const int ret = _vprintf(handle, type, fmt, ap);

	va_end(ap);

	return ret;
}

static int
_dsp_print(void *const data, const char *buf)
{
	plughandle_t *handle = data;

	return _printf(handle, handle->logger.Trace, "%s", buf);
}

static int
_work_print(void *const data, const char *buf)
{
	plughandle_t *handle = data;

	return _printf(handle, handle->logger.Note, "%s", buf);
}

static void
_dsp_assert(void *const data)
{
	jmp_buf *env = data;

	longjmp(*env, 1);
}

static void
_wake_worker(plughandle_t *handle)
{
	const job_t job = {
		.type = JOB_TYPE_WORK
	};

	handle->sched->schedule_work(handle->sched->handle, sizeof(job), &job);
}

static void *
_work_write_request(void *const data, size_t len)
{
	dsp_t *dsp = data;

	return varchunk_write_request(dsp->to_worker, len);
}

static void
_work_write_advance(void *const data, size_t len)
{
	dsp_t *dsp = data;

	varchunk_write_advance(dsp->to_worker, len);
	_wake_worker(dsp->handle);
}

static void const*
_work_read_request(void *const data, size_t *len)
{
	dsp_t *dsp = data;

	return varchunk_read_request(dsp->to_worker, len);
}

static void
_work_read_advance(void *const data)
{
	dsp_t *dsp = data;

	varchunk_read_advance(dsp->to_worker);
}

static void *
_dsp_write_request(void *const data, size_t len)
{
	dsp_t *dsp = data;

	return varchunk_write_request(dsp->to_dsp, len);
}

static void
_dsp_write_advance(void *const data, size_t len)
{
	dsp_t *dsp = data;

	varchunk_write_advance(dsp->to_dsp, len);
}

static void const*
_dsp_read_request(void *const data, size_t *len)
{
	dsp_t *dsp = data;

	return varchunk_read_request(dsp->to_dsp, len);
}

static void
_dsp_read_advance(void *const data)
{
	dsp_t *dsp = data;

	varchunk_read_advance(dsp->to_dsp);
}

static inline voice_t *
_voice_begin(dsp_t *dsp)
{
	return dsp->voices;
}

static inline bool
_voice_not_end(dsp_t *dsp, voice_t *voice)
{
	return (voice - dsp->voices) < JIT_MAX_VOICES;
}

static inline voice_t *
_voice_next(voice_t *voice)
{
	return voice + 1;
}

#define VOICE_FOREACH(DSP, VOICE) \
	for(voice_t *(VOICE) = _voice_begin((DSP)); \
		_voice_not_end((DSP), (VOICE)); \
		(VOICE) = _voice_next((VOICE)))

static void
_intercept_code(void *data, int64_t frames __attribute__((unused)),
	props_impl_t *impl)
{
	plughandle_t *handle = data;

	char *code;
	if( (code = varchunk_write_request(handle->to_loader, impl->value.size)) )
	{
		memcpy(code, handle->state.code, impl->value.size);

		varchunk_write_advance(handle->to_loader, impl->value.size);

		const job_t job = {
			.type = JOB_TYPE_DSP_NEW
		};

		handle->sched->schedule_work(handle->sched->handle, sizeof(job), &job);
	}
	else if(handle->log)
	{
		lv2_log_trace(&handle->logger, "[%s] ringbuffer overflow\n", __func__);
	}
}

static void
_cntrl_refresh_value_abs(float *cntrl, float val)
{
	if(cntrl)
	{
		*cntrl = val;
	}
}

static const props_def_t defs [MAX_NPROPS] = {
	{
		.property = JIT__code,
		.offset = offsetof(plugstate_t, code),
		.type = LV2_ATOM__String,
		.event_cb = _intercept_code,
		.max_size = CODE_SIZE
	},
	{
		.property = JIT__error,
		.access = LV2_PATCH__readable,
		.offset = offsetof(plugstate_t, error),
		.type = LV2_ATOM__String,
		.max_size = ERROR_SIZE
	},
	{
		.property = JIT__xfadeDuration,
		.offset = offsetof(plugstate_t, xfade_dur),
		.type = LV2_ATOM__Int
	},
	{
		.property = JIT__fontHeight,
		.offset = offsetof(plugstate_t, font_height),
		.type = LV2_ATOM__Int
	},
	CONTROL(1),
	CONTROL(2),
	CONTROL(3),
	CONTROL(4),
	CONTROL(5),
	CONTROL(6),
	CONTROL(7),
	CONTROL(1),
	CONTROL(9),
	CONTROL(10),
	CONTROL(11),
	CONTROL(12),
	CONTROL(13),
	CONTROL(14),
	CONTROL(15),
	CONTROL(16)
};

static inline void
_play(plughandle_t *handle, int64_t from, int64_t to)
{
	const uint32_t nsamples = to - from;
	const size_t buflen = nsamples * sizeof(float);

	float *audio_in [3] = {
		alloca(buflen), //FIXME check
		handle->nchannel > 1
			? alloca(buflen)
			: NULL, //FIXME check
		NULL
	};
	float *audio_out [3] = {
		&handle->audio_out[0][from],
		handle->nchannel > 1
			? &handle->audio_out[1][from]
			: NULL,
		NULL
	};

	for(uint32_t n = 0; n < handle->nchannel; n++)
	{
		for(uint32_t i = 0; i < nsamples; i++)
		{
			// clone audio in for in-place compatibility
			audio_in[n][i] = handle->audio_in[n][from + i];

			// generate silence
			audio_out[n][i] = 0.f;
		}
	}

	const bool off [2] = {
		handle->play,
		!handle->play
	};

	for(uint32_t d = 0; d < 2; d++)
	{
		dsp_t *dsp = handle->dsp[off[d]];

		if(!dsp)
		{
			continue;
		}

		float gain = 1.f;

		if(handle->xfade_cur > 0)
		{
			const float t = 2.f * handle->xfade_cur / handle->xfade_max - 1.f;

			if(d == 0) // fade-out
			{
				gain = sqrtf(0.5f * (1.f + t) );
			}
			else // fade-in
			{
				gain = sqrtf(0.5f * (1.f - t) );
			}
		}
		else
		{
			if(d == 1)
			{
				continue; // skip this
			}
		}

		if(dsp->desc && dsp->desc->run)
		{
			jmp_buf env;

			if(setjmp(env) == 0)
			{
				jit_t jit = {
					.mask = JIT_MASK_ALL,
					.srate = handle->srate,
					.nchannels = handle->nchannel,
					.ncontrols = NCONTROLS,
					.nsamples = nsamples,
					.audio_in = (float const *const *)audio_in,
					.audio_out = audio_out,
					.control = handle->state.control,
					._print = {
						.func = _dsp_print,
						.data = handle
					},
					._assert = {
						.func = _dsp_assert,
						.data = &env
					},
					._work_write_request = {
						.func = _work_write_request,
						.data = dsp
					},
					._work_write_advance = {
						.func = _work_write_advance,
						.data = dsp
					},
					._work_read_request = {
						.func = _dsp_read_request,
						.data = dsp
					},
					._work_read_advance = {
						.func = _dsp_read_advance,
						.data = dsp
					}
				};

				dsp->desc->run(&jit, dsp->data);
			}
			else if(handle->log)
			{
				lv2_log_error(&handle->logger, "[%s] assert in run", __func__);
			}
			// add to master out
			for(uint32_t n = 0; n < handle->nchannel; n++)
			{
				for(uint32_t i = 0; i < nsamples; i++)
				{
					handle->audio_out[n][from + i] += gain * audio_out[n][i];
				}
			}
		}
	}

	if(handle->xfade_cur > 0)
	{
		if(nsamples >= handle->xfade_cur)
		{
			handle->xfade_cur = 0;

			// switch dsps
			handle->play = !handle->play;
		}
		else
		{
			handle->xfade_cur -= nsamples;
		}
	}
}

static LV2_Handle
instantiate(const LV2_Descriptor* descriptor, double rate,
	const char *bundle_path, const LV2_Feature *const *features)
{
	plughandle_t *handle = calloc(1, sizeof(plughandle_t));
	if(!handle)
	{
		return NULL;
	}
	mlock(handle, sizeof(plughandle_t));

	handle->nchannel = 1;
	if(!strcmp(descriptor->URI, JIT__stereo))
	{
		handle->nchannel = 2;
	}

	strncpy(handle->bundle_path, bundle_path, sizeof(handle->bundle_path) - 1);

	for(unsigned i=0; features[i]; i++)
	{
		if(!strcmp(features[i]->URI, LV2_URID__map))
		{
			handle->map = features[i]->data;
		}
		else if(!strcmp(features[i]->URI, LV2_WORKER__schedule))
		{
			handle->sched= features[i]->data;
		}
		else if(!strcmp(features[i]->URI, LV2_LOG__log))
		{
			handle->log = features[i]->data;
		}
	}

	if(!handle->map)
	{
		fprintf(stderr,
			"%s: Host does not support urid:map\n", descriptor->URI);
		free(handle);
		return NULL;
	}

	if(!handle->sched)
	{
		fprintf(stderr,
			"%s: Host does not support work:sched\n", descriptor->URI);
		free(handle);
		return NULL;
	}

	if(handle->log)
	{
		lv2_log_logger_init(&handle->logger, handle->map, handle->log);
	}

	lv2_atom_forge_init(&handle->forge, handle->map);

	handle->midi_MidiEvent = handle->map->map(handle->map->handle,
		LV2_MIDI__MidiEvent);

	if(!props_init(&handle->props, descriptor->URI,
		defs, MAX_NPROPS, &handle->state, &handle->stash,
		handle->map, handle))
	{
		fprintf(stderr, "failed to initialize property structure\n");
		free(handle);
		return NULL;
	}

	handle->jit_error = props_map(&handle->props, JIT__error);

	handle->to_loader = varchunk_new(BUF_SIZE, true);
	handle->xfade_max = rate; //FIXME make this configurable
	handle->srate = rate;

	for(uint32_t chn = 0; chn < 0x10; chn++)
	{
		handle->range[chn] = 48.f; // semitones
	}

	return handle;
}

static void
connect_port(LV2_Handle instance, uint32_t port, void *data)
{
	plughandle_t *handle = (plughandle_t *)instance;

	switch(port)
	{
		case 0:
			handle->control = (const LV2_Atom_Sequence *)data;
			break;
		case 1:
			handle->notify = (LV2_Atom_Sequence *)data;
			break;
		case 2:
			handle->audio_in[0] = (const float *)data;
			break;
		case 3:
			handle->audio_out[0] = (float *)data;
			break;
		case 4:
			handle->audio_in[1] = (const float *)data;
			break;
		case 5:
			handle->audio_out[1] = (float *)data;
			break;
		default:
			break;
	}
}

static inline float
_midi2cps(float pitch)
{
	return exp2f( (pitch - 69.f) / 12.f) * 440.f;
}

static inline voice_t *
_next_available_voice(dsp_t *dsp)
{
	VOICE_FOREACH(dsp, voice)
	{
		if(voice->state == VOICE_STATE_INACTIVE)
		{
			return voice;
		}
	}

	VOICE_FOREACH(dsp, voice)
	{
		if(voice->state & VOICE_STATE_DEACTIVATING)
		{
			return voice;
		}
	}

	VOICE_FOREACH(dsp, voice)
	{
		if(voice->state & VOICE_STATE_SUSTAIN)
		{
			return voice;
		}
	}

	return NULL;
}

static inline voice_t *
_find_active_voice(dsp_t *dsp, const hash_t *hash)
{
	VOICE_FOREACH(dsp, voice)
	{
		if(  (voice->state == VOICE_STATE_ACTIVE)
			&& (voice->hash.id == hash->id) )
		{
			return voice;
		}
	}

	return NULL;
}

static inline void
_update_frequency(plughandle_t *handle, dsp_t *dsp, uint8_t chn)
{
	VOICE_FOREACH(dsp, voice)
	{
		if(voice->state & VOICE_STATE_ACTIVE)
		{
			if(voice->hash.chn == chn)
			{
				const float freq = _midi2cps(voice->hash.key
					+ handle->bend[chn]*handle->range[chn]);

				_cntrl_refresh_value_abs(&voice->freq, freq);
			}
		}
	}
}

static inline void
_update_gain(plughandle_t *handle, dsp_t *dsp, uint8_t chn)
{
	VOICE_FOREACH(dsp, voice)
	{
		if(voice->state & VOICE_STATE_ACTIVE)
		{
			if(voice->hash.chn == chn)
			{
				const float gain = handle->gain[chn] * 0x1p-14;

				_cntrl_refresh_value_abs(&voice->gain, gain);
			}
		}
	}
}

static inline void
_update_pressure(plughandle_t *handle, dsp_t *dsp, uint8_t chn)
{
	VOICE_FOREACH(dsp, voice)
	{
		if(voice->state & VOICE_STATE_ACTIVE)
		{
			if(voice->hash.chn == chn)
			{
				const float pressure = handle->pressure[chn] * 0x1p-14;

				_cntrl_refresh_value_abs(&voice->pressure, pressure);
			}
		}
	}
}

static inline void
_update_timbre(plughandle_t *handle, dsp_t *dsp, uint8_t chn)
{
	VOICE_FOREACH(dsp, voice)
	{
		if(voice->state & VOICE_STATE_ACTIVE)
		{
			if(voice->hash.chn == chn)
			{
				const float timbre = handle->timbre[chn] * 0x1p-14;

				_cntrl_refresh_value_abs(&voice->timbre, timbre);
			}
		}
	}
}

static inline void
_voice_off(plughandle_t *handle, voice_t *voice)
{
	if(handle->sustain[voice->hash.chn])
	{
		voice->state |= VOICE_STATE_SUSTAIN;
	}
	else
	{
		_cntrl_refresh_value_abs(&voice->gate, 0.f);

		voice->state |= VOICE_STATE_DEACTIVATING;
		voice->remaining = handle->srate; //FIXME how long ?
	}
}

static inline void
_voice_off_force(voice_t *voice)
{
	voice->state = VOICE_STATE_INACTIVE;
}

static inline void
_voice_send_event(plughandle_t *handle, dsp_t *dsp, jit_ev_type_t etype,
	voice_t *voice)
{
	if(dsp->desc && dsp->desc->event)
	{
		jmp_buf env;

		if(setjmp(env) == 0)
		{
			jit_t jit = {
				.mask = JIT_MASK_ALL
					& ~(JIT_MASK_NSAMPLES | JIT_MASK_AUDIO_IN | JIT_MASK_AUDIO_OUT),
				.srate = handle->srate,
				.nchannels = handle->nchannel,
				.ncontrols = NCONTROLS,
				.control = handle->state.control,
				._print = {
					.func = _dsp_print,
					.data = handle
				},
				._assert = {
					.func = _dsp_assert,
					.data = &env
				},
				._work_write_request = {
					.func = _work_write_request,
					.data = dsp
				},
				._work_write_advance = {
					.func = _work_write_advance,
					.data = dsp
				},
				._work_read_request = {
					.func = _dsp_read_request,
					.data = dsp
				},
				._work_read_advance = {
					.func = _dsp_read_advance,
					.data = dsp
				}
			};

			const jit_ev_t ev = {
				.etype = etype,
				.index = voice - dsp->voices,
				//.gate = voice->gate, FIXME
				.gain = voice->gain,
				.freq = voice->freq,
				.pressure = voice->pressure,
				.timbre = voice->timbre
			};

			dsp->desc->event(&jit, dsp->data, &ev);
		}
		else if(handle->log)
		{
			lv2_log_error(&handle->logger, "[%s] assert in init", __func__);
		}
	}
}

static void
_handle_midi(plughandle_t *handle, dsp_t *dsp,
	int64_t frames __attribute__((unused)), const uint8_t *msg, uint32_t len)
{
	if(len < 3)
	{
		return;
	}

	const uint8_t cmd = msg[0] & 0xf0;
	const uint8_t chn = msg[0] & 0x0f;

	if(!dsp)
	{
		return;
	}

	switch(cmd)
	{
		case LV2_MIDI_MSG_NOTE_ON:
		{
			const uint8_t key = msg[1];
			const uint8_t vel = msg[2];

			voice_t *voice = _next_available_voice(dsp);
			if(voice)
			{
				const float freq = _midi2cps(key
					+ handle->bend[chn]*handle->range[chn]);

				_cntrl_refresh_value_abs(&voice->freq, freq);
				_cntrl_refresh_value_abs(&voice->gain, vel * 0x1p-7);
				_cntrl_refresh_value_abs(&voice->gate, 1.f);

				voice->hash.key = key;
				voice->hash.chn = chn;
				voice->state = VOICE_STATE_ACTIVE;
				voice->remaining = 0;

				_voice_send_event(handle, dsp, JIT_EVENT_ON, voice);
			}
		} break;
		case LV2_MIDI_MSG_NOTE_OFF:
		{
			const uint8_t key = msg[1];

			const hash_t hash = {
				.key= key,
				.chn = chn
			};

			voice_t *voice = _find_active_voice(dsp, &hash);

			if(voice)
			{
				_voice_off(handle, voice);

				_voice_send_event(handle, dsp, JIT_EVENT_OFF, voice);
			}
		} break;
		case LV2_MIDI_MSG_NOTE_PRESSURE:
		{
			const uint8_t key = msg[1];
			const uint8_t pre = msg[2];

			const hash_t hash = {
				.key= key,
				.chn = chn
			};

			voice_t *voice = _find_active_voice(dsp, &hash);

			if(voice)
			{
				_cntrl_refresh_value_abs(&voice->pressure, pre * 0x1p-7);
			}
		} break;
		case LV2_MIDI_MSG_BENDER:
		{
			const uint8_t lsb = msg[1];
			const uint8_t msb = msg[2];
			const int16_t bend = (msb << 7) | lsb;

			handle->bend[chn] = (bend - 0x2000) * 0x1p-13;
			_update_frequency(handle, dsp, chn);
		} break;
		case LV2_MIDI_MSG_CONTROLLER:
		{
			const uint8_t ctr = msg[1];
			const uint8_t val = msg[2];

			switch(ctr)
			{
				case LV2_MIDI_CTL_SUSTAIN:
				{
					//handle->sustain[chn] = (val > 0x3f)
					handle->sustain[chn] = (val < 0x3f)
						? true
						: false;

					if(handle->sustain[chn] == false)
					{
						VOICE_FOREACH(dsp, voice)
						{
							if(  (voice->hash.chn == chn)
								&& (voice->state & VOICE_STATE_SUSTAIN) )
							{
								_voice_off(handle, voice);
							}
						}
					}
				} break;
				case LV2_MIDI_CTL_ALL_NOTES_OFF:
				{
					VOICE_FOREACH(dsp, voice)
					{
						_voice_off(handle, voice);
					}
				} break;
				case LV2_MIDI_CTL_ALL_SOUNDS_OFF:
				{
					VOICE_FOREACH(dsp, voice)
					{
						_voice_off_force(voice);
					}
				} break;
				case LV2_MIDI_CTL_RPN_LSB:
				{
					handle->rpn_lsb[chn] = val;
				} break;
				case LV2_MIDI_CTL_RPN_MSB:
				{
					handle->rpn_msb[chn] = val;
				} break;
				case LV2_MIDI_CTL_LSB_DATA_ENTRY:
				{
					handle->data_lsb[chn] = val;
				} break;
				case LV2_MIDI_CTL_MSB_DATA_ENTRY:
				{
					// pitch-bend range
					if( (handle->rpn_msb[chn] == 0x0) && (handle->rpn_lsb[chn] == 0x0) )
					{
						const uint8_t semi = val;
						const uint8_t cent = handle->data_lsb[chn];

						handle->range[chn] = (float)semi + cent*0.01f;
						_update_frequency(handle, dsp, chn);
					}
				} break;

				case LV2_MIDI_CTL_LSB_MAIN_VOLUME:
				{
					handle->gain[chn] &= 0x7f;
					handle->gain[chn] |= val;
				} break;
				case LV2_MIDI_CTL_MSB_MAIN_VOLUME:
				{
					handle->gain[chn] &= 0x3f80;
					handle->gain[chn] |= val << 7;

					_update_gain(handle, dsp, chn);
				} break;
				case LV2_MIDI_CTL_SC1_SOUND_VARIATION | 0x20:
				{
					handle->pressure[chn] &= 0x7f;
					handle->pressure[chn] |= val;
				} break;
				case LV2_MIDI_CTL_SC1_SOUND_VARIATION:
				{
					handle->pressure[chn] &= 0x3f80;
					handle->pressure[chn] |= val << 7;

					_update_pressure(handle, dsp, chn);
				} break;
				case LV2_MIDI_CTL_SC5_BRIGHTNESS | 0x20:
				{
					handle->timbre[chn] &= 0x7f;
					handle->timbre[chn] |= val;
				} break;
				case LV2_MIDI_CTL_SC5_BRIGHTNESS:
				{
					handle->timbre[chn] &= 0x3f80;
					handle->timbre[chn] |= val << 7;

					_update_timbre(handle, dsp, chn);
				} break;
			}
		} break;
	}
}

static void
run(LV2_Handle instance, uint32_t nsamples)
{
	plughandle_t *handle = instance;

	const uint32_t capacity = handle->notify->atom.size;
	LV2_Atom_Forge_Frame frame;
	lv2_atom_forge_set_buffer(&handle->forge, (uint8_t *)handle->notify, capacity);
	handle->ref = lv2_atom_forge_sequence_head(&handle->forge, &frame, 0);

	props_idle(&handle->props, &handle->forge, 0, &handle->ref);

	int64_t last_t = 0;
	LV2_ATOM_SEQUENCE_FOREACH(handle->control, ev)
	{
		const LV2_Atom *atom = &ev->body;
		const LV2_Atom_Object *obj = (const LV2_Atom_Object *)&ev->body;

		if(atom->type == handle->midi_MidiEvent)
		{
			const bool off [2] = {
				handle->play,
				!handle->play
			};

			for(uint32_t d = 0; d < 2; d++)
			{
				dsp_t *dsp = handle->dsp[off[d]];

				if(!dsp)
				{
					continue;
				}

				if(handle->xfade_cur > 0)
				{
					// do nothing
				}
				else
				{
					if(d == 1)
					{
						continue; // skip this
					}
				}

				_handle_midi(handle, dsp, ev->time.frames, LV2_ATOM_BODY_CONST(atom),
					atom->size);
			}
		}
		else
		{
			props_advance(&handle->props, &handle->forge, ev->time.frames, obj,
				&handle->ref);
		}

		_play(handle, last_t, ev->time.frames);

		last_t = ev->time.frames;
	}

	_play(handle, last_t, nsamples);

	// send error if applicable
	if(handle->dirty)
	{
		props_set(&handle->props, &handle->forge, nsamples-1, handle->jit_error,
			&handle->ref);

		handle->dirty = false;
	}

	if(handle->ref)
	{
		lv2_atom_forge_pop(&handle->forge, &frame);
	}
	else
	{
		lv2_atom_sequence_clear(handle->notify);

		if(handle->log)
		{
			lv2_log_trace(&handle->logger, "forge buffer overflow\n");
		}
	}
}

static uint64_t
_hash_file(plughandle_t *handle, const char *file, uint64_t seed)
{
	uint64_t hash = 0;
	char *path = NULL;

	if(asprintf(&path, "%s%s", handle->bundle_path, file) == -1)
	{
		return hash;
	}

	// load file
	FILE *f = fopen(path, "rb");
	if(f)
	{
		fseek(f, 0, SEEK_END);
		long fsize = ftell(f);
		fseek(f, 0, SEEK_SET);

		char *str = malloc(fsize);
		if(str)
		{
			if(fread(str, fsize, 1, f) == 1)
			{
				hash = mum_hash(str, fsize, seed);
			}

			free(str);
		}

		fclose(f);
	}

	free(path);

	return hash;
}

static dsp_t *
_dsp_new(plughandle_t *handle, const char *code, size_t size,
	LV2_Worker_Respond_Function respond, LV2_Worker_Respond_Handle target)
{
	dsp_t *dsp = calloc(1, sizeof(dsp_t));

	if(!dsp)
	{
		return NULL;
	}

	dsp->handle = handle;
	dsp->to_worker = varchunk_new(BUF_SIZE, true); //FIXME check
	dsp->to_dsp = varchunk_new(BUF_SIZE, true); //FIXME check

	uint64_t hash = 12345; // seed
	hash = _hash_file(handle, "jit.h", hash);
	hash = _hash_file(handle, "jit.rs", hash);
	hash = mum_hash(code, size, hash);

	snprintf(dsp->template_so, sizeof(dsp->template_so),
		"/tmp/jit_%016"PRIx64".so", hash);

	{
		const job_t job = {
			.type = JOB_TYPE_ERROR_CLEAR,
			.error = NULL
		};

		respond(target, sizeof(job), &job);
	}

	if(access(dsp->template_so, F_OK) == 0)
	{
		if(handle->log)
		{
			lv2_log_note(&handle->logger, "using cached object '%s'", dsp->template_so);
		}
	}
	else
	{
		if(handle->log)
		{
			lv2_log_note(&handle->logger, "compiling new object'%s'", dsp->template_so);
		}

		snprintf(dsp->template_src, sizeof(dsp->template_src),
			"/tmp/jit_%016"PRIx64, hash);

		FILE *fout = fopen(dsp->template_src, "wb");
		if(!fout)
		{
			goto fail_errno;
		}

		if(fwrite(code, size - 1, 1, fout) != 1)
		{
			fclose(fout);
			unlink(dsp->template_src);
			goto fail_errno;
		}
		fclose(fout);

		regex_t re;
		if( (errno = regcomp(&re, "jit:(.*):$", REG_EXTENDED | REG_NEWLINE) ) != 0)
		{
			unlink(dsp->template_src);
			regfree(&re);
			goto fail_errno;
		}

		regmatch_t pmatch [2];
		memset(&pmatch, 0x0, sizeof(pmatch));

		if(regexec(&re, code, 2, pmatch, 0) != 0)
		{
			unlink(dsp->template_src);
			regfree(&re);
			goto fail_errno;
		}
		regfree(&re);

		if( (pmatch[1].rm_so == -1) || (pmatch[1].rm_eo == -1) )
		{
			unlink(dsp->template_src);
			goto fail_errno; //FIXME set errno
		}

		char *fmt = strndup(&code[pmatch[1].rm_so], pmatch[1].rm_eo - pmatch[1].rm_so);
		if(!fmt)
		{
			unlink(dsp->template_src);
			goto fail_errno; //FIXME set errno
		}

		const char *arr [3] = { NULL, NULL, NULL };
		int num = 0;
		for(char *ptr = strchr(fmt, '%'); ptr && (num < 3); ptr = strchr(&ptr[1], '%'))
		{
			switch(ptr[1])
			{
				case 'b':
				{
					ptr[1] = 's';
					arr[num++] = handle->bundle_path;
				} break;
				case 'o':
				{
					ptr[1] = 's';
					arr[num++] = dsp->template_so;
				} break;
				case 'i':
				{
					ptr[1] = 's';
					arr[num++] = dsp->template_src;
				} break;
				default:
				{
					free(fmt);
					unlink(dsp->template_src);
					goto fail_errno; //FIXME set errno
				} break;
			}
		}

		char *cmd = NULL;
		if(asprintf(&cmd, fmt, arr[0], arr[1], arr[2]) == -1)
		{
			free(fmt);
			unlink(dsp->template_src);
			goto fail_errno; //FIXME set errno
		}
		free(fmt);

		FILE *fin = fin = popen(cmd, "r");
		if(!fin)
		{
			unlink(dsp->template_src);
			free(cmd);
			goto fail_errno;
		}
		free(cmd);

		char *ln = NULL;
		size_t len = 0;
		while(getline(&ln, &len, fin) != -1)
		{
			if(handle->log)
			{
				lv2_log_error(&handle->logger, "%s", ln);
			}

			const job_t job = {
				.type = JOB_TYPE_ERROR_APPEND,
				.error = strdup(ln)
			};

			respond(target, sizeof(job), &job);
		}
		unlink(dsp->template_src);

		errno = pclose(fin);
		if(errno != 0)
		{
			goto fail_errno;
		}
	}

	// load compiled code
	dsp->lib = dlopen(dsp->template_so, RTLD_NOW | RTLD_LOCAL);
	if(!dsp->lib)
	{
		goto fail_dlerror;
	}

	dsp->desc = (jit_desc_t const *)dlsym(dsp->lib, "jit_desc");
	if(!dsp->desc)
	{
		dlclose(dsp->lib);
		goto fail_dlerror;
	}

	if(dsp->desc && dsp->desc->init)
	{
		jmp_buf env;

		if(setjmp(env) == 0)
		{
			jit_t jit = {
				.mask = JIT_MASK_ALL
					& ~(JIT_MASK_NSAMPLES | JIT_MASK_AUDIO_IN | JIT_MASK_AUDIO_OUT),
				.srate = handle->srate,
				.nchannels = handle->nchannel,
				.ncontrols = NCONTROLS,
				.control = handle->stash.control,
				._print = {
					.func = _work_print,
					.data = handle
				},
				._assert = {
					.func = _dsp_assert,
					.data = &env
				},
				._work_write_request = {
					.func = _dsp_write_request,
					.data = dsp
				},
				._work_write_advance = {
					.func = _dsp_write_advance,
					.data = dsp
				},
				._work_read_request = {
					.func = _work_read_request,
					.data = dsp
				},
				._work_read_advance = {
					.func = _work_read_advance,
					.data = dsp
				}
			};

			dsp->data = dsp->desc->init(&jit);
		}
		else
		{
			goto fail_init;
		}
	}

	if(handle->log)
	{
		lv2_log_note(&handle->logger, "[%s] loading succeeded: %s", __func__,
			dsp->template_so);
	}

	return dsp;

fail_errno:
	if(handle->log)
	{
		lv2_log_error(&handle->logger, "[%s] %s\n", __func__, strerror(errno));
	}

	free(dsp);
	return NULL;

fail_dlerror:
	if(handle->log)
	{
		lv2_log_error(&handle->logger, "[%s] %s\n", __func__, dlerror());
	}

	free(dsp);
	return NULL;

fail_init:
	if(handle->log)
	{
			lv2_log_error(&handle->logger, "[%s] assert in init", __func__);
	}

	free(dsp);
	return NULL;
}

static void
_dsp_free(plughandle_t *handle, dsp_t *dsp)
{
	if(dsp)
	{
		if(dsp->lib)
		{
			if(dsp->desc && dsp->desc->deinit)
			{
				jmp_buf env;

				if(setjmp(env) == 0)
				{
					jit_t jit = {
						.mask = JIT_MASK_ALL
							& ~(JIT_MASK_NSAMPLES | JIT_MASK_AUDIO_IN | JIT_MASK_AUDIO_OUT),
						.srate = handle->srate,
						.nchannels = handle->nchannel,
						.ncontrols = NCONTROLS,
						.control = handle->stash.control,
						._print = {
							.func = _work_print,
							.data = handle
						},
						._assert = {
							.func = _dsp_assert,
							.data = &env
						},
						._work_write_request = {
							.func = _dsp_write_request,
							.data = dsp
						},
						._work_write_advance = {
							.func = _dsp_write_advance,
							.data = dsp
						},
						._work_read_request = {
							.func = _work_read_request,
							.data = dsp
						},
						._work_read_advance = {
							.func = _work_read_advance,
							.data = dsp
						}
					};

					dsp->desc->deinit(&jit, dsp->data);
				}
				else if(handle->log)
				{
					lv2_log_error(&handle->logger, "[%s] assert in deinit", __func__);
				}
			}

			dlclose(dsp->lib);
			//unlink(dsp->template_so);
		}

		varchunk_free(dsp->to_worker);
		varchunk_free(dsp->to_dsp);

		free(dsp);
	}
	else if(handle->log)
	{
		lv2_log_note(&handle->logger, "invalid dsp pointer");
	}
}

static void
cleanup(LV2_Handle instance)
{
	plughandle_t *handle = instance;

	munlock(handle, sizeof(plughandle_t));
	varchunk_free(handle->to_loader);
	_dsp_free(handle, handle->dsp[0]);
	_dsp_free(handle, handle->dsp[1]);
	free(handle);
}

static LV2_State_Status
_state_save(LV2_Handle instance, LV2_State_Store_Function store,
	LV2_State_Handle state, uint32_t flags,
	const LV2_Feature *const *features)
{
	plughandle_t *handle = instance;

	return props_save(&handle->props, store, state, flags, features);
}

static LV2_State_Status
_state_restore(LV2_Handle instance, LV2_State_Retrieve_Function retrieve,
	LV2_State_Handle state, uint32_t flags,
	const LV2_Feature *const *features)
{
	plughandle_t *handle = instance;

	return props_restore(&handle->props, retrieve, state, flags, features);
}

static const LV2_State_Interface state_iface = {
	.save = _state_save,
	.restore = _state_restore
};

// non-rt thread
static LV2_Worker_Status
_work(LV2_Handle instance,
	LV2_Worker_Respond_Function respond,
	LV2_Worker_Respond_Handle target,
	uint32_t body_size,
	const void *body)
{
	plughandle_t *handle = instance;

	if(body_size != sizeof(job_t))
	{
		return LV2_WORKER_ERR_UNKNOWN;
	}

	const job_t *job = body;
	switch(job->type)
	{
		case JOB_TYPE_DSP_NEW:
		{
			size_t size;
			const char *code;
			while( (code= varchunk_read_request(handle->to_loader, &size)) )
			{
				dsp_t *dsp = _dsp_new(handle, code, size, respond, target);
				if(dsp)
				{
					const job_t job2 = {
						.type = JOB_TYPE_DSP_SET,
						.dsp = dsp
					};

					respond(target, sizeof(job2), &job2);
				}

				varchunk_read_advance(handle->to_loader);
			}
		} break;
		case JOB_TYPE_DSP_SET:
		{
			// never reached
		} break;
		case JOB_TYPE_DSP_FREE:
		{
			_dsp_free(handle, job->dsp);
		} break;
		case JOB_TYPE_ERROR_CLEAR:
		{
			// never reached
		} break;
		case JOB_TYPE_ERROR_APPEND:
		{
			// never reached
		} break;
		case JOB_TYPE_ERROR_FREE:
		{
			if(job->error)
			{
				free(job->error);
			}
		} break;
		case JOB_TYPE_WORK:
		{
			dsp_t *dsp = handle->dsp[handle->play];

			if(dsp && dsp->desc && dsp->desc->work)
			{
				jmp_buf env;

				if(setjmp(env) == 0)
				{
					jit_t jit = {
						.mask = JIT_MASK_ALL
							& ~(JIT_MASK_NSAMPLES | JIT_MASK_AUDIO_IN | JIT_MASK_AUDIO_OUT),
						.srate = handle->srate,
						.nchannels = handle->nchannel,
						.ncontrols = NCONTROLS,
						.control = handle->stash.control,
						._print = {
							.func = _work_print,
							.data = handle
						},
						._assert = {
							.func = _dsp_assert,
							.data = &env
						},
						._work_write_request = {
							.func = _dsp_write_request,
							.data = dsp
						},
						._work_write_advance = {
							.func = _dsp_write_advance,
							.data = dsp
						},
						._work_read_request = {
							.func = _work_read_request,
							.data = dsp
						},
						._work_read_advance = {
							.func = _work_read_advance,
							.data = dsp
						}
					};

					dsp->desc->work(&jit, dsp->data);
				}
				else
				{
					//FIXME
				}
			}
		} break;
		default:
		{
			// never reached
		} return LV2_WORKER_ERR_UNKNOWN;
	}

	return LV2_WORKER_SUCCESS;
}

// rt-thread
static LV2_Worker_Status
_work_response(LV2_Handle instance, uint32_t size, const void *body)
{
	plughandle_t *handle = instance;

	if(size != sizeof(job_t))
	{
		return LV2_WORKER_ERR_UNKNOWN;
	}

	const job_t *job = body;

	switch(job->type)
	{
		case JOB_TYPE_DSP_NEW:
		{
			// never reached
		} break;
		case JOB_TYPE_DSP_SET:
		{
			const job_t job2 = {
				.type = JOB_TYPE_DSP_FREE,
				.dsp = handle->dsp[!handle->play]
			};

			handle->sched->schedule_work(handle->sched->handle, sizeof(job2), &job2);

			handle->dsp[!handle->play] = job->dsp;
			handle->xfade_cur = handle->xfade_max;
		} break;
		case JOB_TYPE_DSP_FREE:
		{
			// never reached
		} break;
		case JOB_TYPE_ERROR_CLEAR:
		{
			props_impl_t *impl = _props_impl_get(&handle->props, handle->jit_error);

			if(impl)
			{
				static const char *empty = "";

				strncpy(handle->state.error, empty, impl->def->max_size);
				impl->value.size = strlen(handle->state.error) + 1;

				handle->dirty = true;
			}
		} break;
		case JOB_TYPE_ERROR_APPEND:
		{
			props_impl_t *impl = _props_impl_get(&handle->props, handle->jit_error);

			if(impl)
			{
				strncat(handle->state.error, job->error, impl->def->max_size);
				impl->value.size = strlen(handle->state.error) + 1;

				handle->dirty = true;

				const job_t job2 = {
					.type = JOB_TYPE_ERROR_FREE,
					.error = job->error
				};

				handle->sched->schedule_work(handle->sched->handle, sizeof(job2), &job2);
			}
		} break;
		case JOB_TYPE_ERROR_FREE:
		{
			// never reached
		} break;
		case JOB_TYPE_WORK:
		{
			// never reached
		} break;
		default:
		{
			// never reached
		} return LV2_WORKER_ERR_UNKNOWN;
	}

	return LV2_WORKER_SUCCESS;
}

// rt-thread
static LV2_Worker_Status
_end_run(LV2_Handle instance __attribute__((unused)))
{
	// do nothing

	return LV2_WORKER_SUCCESS;
}

static const LV2_Worker_Interface work_iface = {
	.work = _work,
	.work_response = _work_response,
	.end_run = _end_run
};

static const void*
extension_data(const char* uri)
{
	if(!strcmp(uri, LV2_STATE__interface))
	{
		return &state_iface;
	}
	else if(!strcmp(uri, LV2_WORKER__interface))
	{
		return &work_iface;
	}

	return NULL;
}

static const LV2_Descriptor jit_mono = {
	.URI						= JIT__mono,
	.instantiate		= instantiate,
	.connect_port		= connect_port,
	.activate				= NULL,
	.run						= run,
	.deactivate			= NULL,
	.cleanup				= cleanup,
	.extension_data	= extension_data
};

static const LV2_Descriptor jit_stereo = {
	.URI						= JIT__stereo,
	.instantiate		= instantiate,
	.connect_port		= connect_port,
	.activate				= NULL,
	.run						= run,
	.deactivate			= NULL,
	.cleanup				= cleanup,
	.extension_data	= extension_data
};

LV2_SYMBOL_EXPORT const LV2_Descriptor*
lv2_descriptor(uint32_t index)
{
	switch(index)
	{
		case 0:
			return &jit_mono;
		case 1:
			return &jit_stereo;
		default:
			return NULL;
	}
}
