extern crate jit;

use jit::{api, util, trace};
use std::ffi;

#[derive(Debug)]
struct Plug {
    count: u32
}

fn wrap_plug<'a>(plug: *mut ffi::c_void)
    -> &'a mut Plug
{
    unsafe {
        &mut *(plug as *mut Plug)
    }
}

extern fn init(jit: *const api::Jit)
    -> *mut ffi::c_void
{
    trace!(jit, "{}", "init");

    let plug = Box::new(Plug {
        count: 0
    });

    Box::into_raw(plug) as *mut ffi::c_void
}

extern fn deinit(jit: *const api::Jit, plug: *mut ffi::c_void)
{
    trace!(jit, "{}", "deinit");

    let plug = wrap_plug(plug);
    let plug = unsafe {
        Box::from_raw(plug)
    };

    drop(plug);
}

fn wake_worker(jit: *const api::Jit, plug: &mut Plug)
{
    match util::work_write_request::<u32>(jit, 1) {
        None => (),
        Some(arr) => {
            for val in arr.iter_mut() { *val = plug.count }

            util::work_write_advance::<u32>(jit, arr.len())
        }
    }
}

extern fn run(jit: *const api::Jit, plug: *mut ffi::c_void)
{
    let plug = wrap_plug(plug);

    let control = match util::control(jit) {
        None => return util::assert(jit, false),
        Some(v) => v
    };
    let audio_in_channels = match util::audio_in_channels(jit) {
        None => return util::assert(jit, false),
        Some(v) => v
    };
    let audio_out_channels = match util::audio_out_channels(jit) {
        None => return util::assert(jit, false),
        Some(v) => v
    };
    let _gain = control[0];

    for audio_out_channel in audio_out_channels.iter()
    {
        let _audio_out = util::audio_out(jit, *audio_out_channel);
    }

    for audio_in_channel in audio_in_channels.iter()
    {
        let _audio_in = util::audio_in(jit, *audio_in_channel);
    }

    plug.count += 1;

    wake_worker(jit, plug)
}

fn drain_worker(jit: *const api::Jit)
{
    match util::work_read_request::<u32>(jit) {
        None => (),
        Some(arr) => {
            for val in arr.iter() { trace!(jit, "drain: {}", val) }

            util::work_read_advance::<u32>(jit)
        }
    }
}

extern fn event(_jit: *const api::Jit, _plug: *mut ffi::c_void,
    _ev: *const api::JitEv)
{
    // nothing to do
}

extern fn work(jit: *const api::Jit, _plug: *mut ffi::c_void)
{
    trace!(jit, "{}", "work");

    drain_worker(jit)
}

#[no_mangle]
pub static jit_desc: api::JitDesc = api::JitDesc {
    init: init,
    deinit: deinit,
    run: run,
    event : event,
    work: work
};

// vim: set syntax=rust:
// jit: rustc --crate-type=cdylib -C opt-level=3 -L crate=%b -o %o %i 2>&1:
