#include <stddef.h>
#include <stdlib.h>
#include <inttypes.h>
#include <math.h>

#include <jit_api.h>
#include <jit_util.h>

#define MAX_SAMPLES 0x10000
#define MASK_SAMPLES (MAX_SAMPLES - 1)

typedef struct _voice_t voice_t;
typedef struct _plug_t plug_t;

typedef enum _state_t {
	STATE_INACTIVE = 0,
	STATE_RAMP_UP,
	STATE_ACTIVE,
	STATE_RAMP_DOWN
} state_t;

struct _voice_t {
	jit_ev_t ev;	
	state_t state;
	float offset;
	float step;
	uint32_t sps;
};

struct _plug_t {
	voice_t voices [JIT_MAX_VOICES];
	float step;
	float wave [MAX_SAMPLES];
	uint32_t sps;
	float _sps;
};

static void *
init(jit_t *const jit)
{
	jit_printf(jit, "[%s]", __func__);

	plug_t *plug = calloc(1, sizeof(plug_t));
	jit_assert(jit, plug);

	for(unsigned i = 0; i < MAX_SAMPLES; i++)
	{
		plug->wave[i] = sinf(i*2.f*M_PI/MASK_SAMPLES) * 0.2f;
	}

	plug->step = MAX_SAMPLES / jit->srate;
	plug->sps = jit->srate * 0.01;
	plug->_sps = 1.f / plug->sps;

	return plug;
}

static void
deinit(jit_t *const jit, void *const data)
{
	jit_printf(jit, "[%s]", __func__);

	plug_t *plug = data;
	jit_assert(jit, plug);

	free(plug);
}

static void
run(jit_t *const jit, void *const data)
{
	plug_t *plug = data;
	jit_assert(jit, plug);

	for(uint32_t n = 0; n < jit->nchannels; n++)
	{
		for(uint32_t i = 0; i < jit->nsamples; i++)
		{
			jit->audio_out[n][i] = 0.f;
		}
	}

	for(unsigned k = 0; k < JIT_MAX_VOICES; k++)
	{
		voice_t *voice = &plug->voices[k];

		if(voice->state == STATE_INACTIVE)
		{
			continue;
		}

		for(uint32_t i = 0; i < jit->nsamples; i++)
		{
			const uint32_t offset = voice->offset;
			float gain;

			switch(voice->state)
			{
				case STATE_RAMP_UP:
				{
					gain = 1.f - voice->sps * plug->_sps;

					if(--voice->sps == 0)
					{
						voice->state = STATE_ACTIVE;
					}
				} break;
				case STATE_RAMP_DOWN:
				{
					gain = voice->sps * plug->_sps;

					if(--voice->sps == 0)
					{
						voice->state = STATE_INACTIVE;
					}
				} break;
				default:
				{
					gain = 1.f;
				} break;
			}

			for(uint32_t n = 0; n < jit->nchannels; n++)
			{
				jit->audio_out[n][i] += plug->wave[offset] * gain;
			}

			voice->offset += voice->step;

			float offset_i;
			const float offset_f = modff(voice->offset, &offset_i);
			voice->offset = ((uint32_t)offset_i & MASK_SAMPLES) + offset_f;
		}
	}
}

static void
on(jit_t *const jit, plug_t *plug, jit_ev_t const *const ev)
{
	jit_printf(jit, "[%s] %"PRIu32" %f", __func__, ev->index, ev->freq);

	voice_t *voice = &plug->voices[ev->index];
	voice->ev = *ev;
	voice->state= STATE_RAMP_UP;
	voice->sps = plug->sps;
	voice->offset = 0.f;
	voice->step = voice->ev.freq * plug->step;
}

static void
off(jit_t *const jit, plug_t *plug, jit_ev_t const *const ev)
{
	jit_printf(jit, "[%s] %"PRIu32, __func__, ev->index);

	voice_t *voice = &plug->voices[ev->index];
	voice->ev = *ev;
	voice->state = STATE_RAMP_DOWN;
	voice->sps = plug->sps;
}

static void
set(jit_t *const jit, plug_t *plug, jit_ev_t const *const ev)
{
	jit_printf(jit, "[%s] %"PRIu32, __func__, ev->index);

	voice_t *voice = &plug->voices[ev->index];
	voice->ev = *ev;
	voice->step = voice->ev.freq * plug->step;
}

static void
event(jit_t *const jit, void *data, jit_ev_t const *const ev)
{
	plug_t *plug = data;
	jit_assert(jit, plug);

	const jit_ev_type_t etype = ev->etype;

	switch(etype)
	{
		case JIT_EVENT_ON:
		{
			on(jit, plug, ev);
		} break;
		case JIT_EVENT_OFF:
		{
			off(jit, plug, ev);
		} break;
		case JIT_EVENT_SET:
		{
			set(jit, plug, ev);
		} break;
	}
}

JIT_EXPORT jit_desc_t const jit_desc = {
	.init = init,
	.deinit = deinit,
	.run = run,
	.event = event
};

// vim: set syntax=c:
// jit: gcc -std=gnu11 -ffast-math -fvisibility=hidden -O3 -shared -fpic -fPIC -Wall -Wextra -Wpedantic -Wl,-z,defs -I %b -o %o -xc %i -lm 2>&1:
