use api;

use std::mem;
use std::fmt;
use std::slice;
use std::os::raw;
use std::io::Write;

fn wrap_jit<'a>(jit: *const api::Jit)
    -> &'a api::Jit
{
    unsafe {
        &*jit
    }
}

pub fn control<'a>(jit: *const api::Jit)
    -> Option<&'a [f32]>
{
    let jit = wrap_jit(jit);

    if jit.control.is_null() {
        None
    } else {
        Some( unsafe {
            slice::from_raw_parts(jit.control, jit.ncontrols as usize)
        } )
    }
}

pub fn audio_in_channels<'a>(jit: *const api::Jit)
    -> Option<&'a [*const f32]>
{
    let jit = wrap_jit(jit);

    if jit.audio_in.is_null() {
        None
    } else {
        Some( unsafe {
            slice::from_raw_parts(jit.audio_in, jit.nchannels as usize)
        } )
    }
}

pub fn audio_out_channels<'a>(jit: *const api::Jit)
    -> Option<&'a [*mut f32]>
{
    let jit = wrap_jit(jit);

    if jit.audio_out.is_null() {
        None
    } else {
        Some( unsafe {
            slice::from_raw_parts(jit.audio_out, jit.nchannels as usize)
        } )
    }
}

pub fn audio_in<'a>(jit: *const api::Jit, channel: *const f32)
    -> &'a [f32]
{
    let jit = wrap_jit(jit);

    unsafe { slice::from_raw_parts(channel, jit.nsamples as usize) }
}

pub fn audio_out<'a>(jit: *const api::Jit, channel: *mut f32)
    -> &'a mut [f32]
{
    let jit = wrap_jit(jit);

    unsafe { slice::from_raw_parts_mut(channel, jit.nsamples as usize) }
}

pub fn assert(jit: *const api::Jit, condition: bool)
{
	let jit = wrap_jit(jit);

    (jit.assert.func)(jit.assert.data, condition)
}

pub fn print(jit: *const api::Jit, fmt: fmt::Arguments)
{
	let jit = wrap_jit(jit);
    let mut vec = [0u8; 512];

    match write!(&mut vec[..], "{}", fmt) {
        Err(_e) => return,
        Ok(_v) => {
            (jit.print.func)(jit.print.data, vec.as_ptr() as *const raw::c_char)
        }
    }
}

pub fn work_write_request<'a, T>(jit: *const api::Jit, nelem: usize)
    -> Option<&'a mut [T]>
{
	let jit = wrap_jit(jit);
    let size = nelem * mem::size_of::<T>();
    let data  =
        (jit.work_write_request.func)(jit.work_write_request.data, size)
        as *mut T;

    if data.is_null() {
        None
    } else {
        Some( unsafe { slice::from_raw_parts_mut(data, nelem) } )
    }
}

pub fn work_write_advance<T>(jit: *const api::Jit, nelem: usize)
{
	let jit = wrap_jit(jit);
    let size = nelem * mem::size_of::<T>();

    (jit.work_write_advance.func)(jit.work_write_advance.data, size);
}

pub fn work_read_request<'a, T>(jit: *const api::Jit)
    -> Option<&'a [T]>
{
	let jit = wrap_jit(jit);
    let mut size = 0usize;
    let data  =
        (jit.work_read_request.func)(jit.work_read_request.data, &mut size)
        as *const T;
    let nelem = size / mem::size_of::<T>();

    if data.is_null() {
        None
    } else {
        Some( unsafe { slice::from_raw_parts(data, nelem) } )
    }
}

pub fn work_read_advance<T>(jit: *const api::Jit)
{
	let jit = wrap_jit(jit);

    (jit.work_read_advance.func)(jit.work_read_advance.data);
}
